/*
  automatic integration by the DE rule 
  for an integral over the infinite interval (0,+infinity)
  whose integrand is of the form 
  f(x) = f0(x)exp(-x)
*/
#define PRECISION 100
#include <iostream>
#include <iomanip>
#include <math.h>
#include <complex>
#include <exfloat.h>
using namespace std;

exfloat ZERO = 0, ONE = 1, TWO = 2;
exfloat HALF = "0.5", QUARTER = "0.25";
exfloat PI = "#pi";
complex<exfloat> I(0, 1);
//
// the DE transform
//
exfloat psi(exfloat u)
{
  return exp(u - exp(-u));
}
//
// the derivative of the DE transform
//
exfloat dpsi(exfloat u)
{
  exfloat ex = exp(-u);
  return (ONE + ex) * exp(u - ex);
}
//
// function for automatic integration, 
// which returns the integral
// 
// (input) complex<exfloat> (*func)(exfloat)
//         ... the integrand which is propotional to exp(-x)
// (output) int n_intgrl
//         ... the number of functional evaluations
//
complex<exfloat> integral(complex<exfloat> (*func)(exfloat), 
			  int & n_intgrl)
{
  cout << setiosflags(ios::scientific);
  //
  exfloat eps = "1.0e-100";  
  // small positive number used for the truncation of the trapezoidal rule
  exfloat h = ONE;  // mesh size of the trapezoidal rule
  n_intgrl = 0;
  //
  // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  // 
  // initial step of the automatic integration
  // with mesh size h = 1
  //
  complex<exfloat> s, term;
  //
  s = ZERO;
  int kmax = 5000;
  //
  // the trapezoidal rule on the positive side
  //
  int l_conv = 0;
  for (int k=0; k<=kmax; ++k)
    {
      exfloat u, x;
      u = k * h;
      x = psi(u);
      term = func(psi(u)) * dpsi(u);
      ++n_intgrl;
      s += term;
      if (abs(term) < eps) {   // if |term| < eps, truncate the infinite sum
	++l_conv;              // of the trapezoidal rule
	break;
      }
    }
  if (l_conv==0) 
    cout << "DE rule not convergent on the positive side." << endl;
  //
  // the trapezoidal rule on the negative side
  //
  l_conv = 0;
  for (int k=1; k<=kmax; ++k)
    {
      exfloat u, x;
      u = - k * h;
      x = psi(u);
      term = func(psi(u)) * dpsi(u);
      ++n_intgrl;
      s += term;
      if (abs(term) < eps) {   // if |term| < eps, truncate the infinite sum
	++l_conv;              // of the trapezoidal rule
	break;
      }
    }
  if (l_conv==0)
    cout << "DE rule not convergent on the negative side." << endl;
  //
  s *= h;
  /*
  cout << setprecision(10) << h << " " 
       << setw(5) << n_intgrl << " "
       << setprecision(15) << s << endl;
  */
  //
  //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  //
  // general step of the automatic integration
  //
  exfloat epsq = sqrt(eps), c_safe = "0.2";
  complex<exfloat> s_old;
  int l_conv1 = 0;
  int lmax = 7;
  for (int l=1; l<=lmax; ++l)
    {
      h *= HALF;
      s_old = s;   // s_old = I(2h) : integral with mesh size 2h
      s = ZERO;    // s     = I( h) : integral with mesh size  h 
      // 
      // the trapezoidal rule on the positive side
      // 
      l_conv = 0;
      for (int k=1; k<=kmax; k+=2)
	{
	  exfloat u, x;
	  u = k * h;
	  x = psi(u);
	  term = func(psi(u)) * dpsi(u);
	  ++n_intgrl;
	  s += term;
	  if (abs(term) < eps) { // if |term| < eps, truncate the infinite sum
	    ++l_conv;            // of the trapezoidal rule
	    break;
	  }
	}
      if (l_conv==0) 
	cout << "DE rule not convergent on the positive side." << endl;
      //
      // the trapezoidal rule on the negative side
      //
      l_conv = 0;
      for (int k=1; k<=kmax; k+=2)
	{
	  exfloat u, x;
	  u = - k * h;
	  x = psi(u);
	  term = func(psi(u)) * dpsi(u);
	  ++n_intgrl;
	  s += term;
	  if (abs(term) < eps) { // if |term| < eps, truncate the infinite sum
	    ++l_conv;            // of the trapezoidal rule
	    break;
	  }
	}
      if (l_conv==0)
	cout << "DE rule not convergent on the negative side." << endl;
      //
      s = h * s + HALF * s_old;
      /*
      cout << setprecision(10) << h << " " 
	   << setw(5) << n_intgrl << " "
	   << setprecision(15) << s << endl;
      */
      //
      if (abs(s - s_old) < c_safe * epsq) {
	++l_conv1; // if |I(h) - I(2h)| < c_safe * eps^{1/2}, 
	break;     // stop the iterations of the automatic integration
      }            // and return I(h) as the approximate integral
    }
  if (l_conv1==0) 
    cout << "automatic integration not convergent." << endl;
  //
  return s;
}
//
// example of the integrand function
//
complex<exfloat> func(exfloat x)
{
  return exp(-(ONE + I)*x);
}

int main(void)
{
  cout << setiosflags(ios::scientific);
  //
  complex<exfloat> s, se;
  // s  : approximate integral by the DE rule
  // se : exact value of the integral 
  exfloat error; // relative error of the approximate integral
  int n_intgrl;  // number of the functional evaluations
  //                in the numerical integration
  //
  se = ONE / (ONE + I);
  s = integral(func, n_intgrl);
  error = abs(s - se) / abs(se);
  cout << "integral(exact)    : " << setprecision(15) << se << endl;
  cout << "integral(numerical): " << setprecision(15) << s << endl;	
  cout << "relarive error     : " << setprecision(1) << error << endl;
  cout << "number of functional evaluations: " << setw(5) << n_intgrl << endl;
  //
  return 0;
}
